#include <stdio.h>
#include <stdlib.h>
#include <unistd.h>
#define _GNU_SOURCE
#include <getopt.h>
#include "plugin.h"
#include "read_config.h"
#include "main_loop.h"

#define CONFIG_FILE "/audio_visualization.conf"
#define RELATIVE_FIFO_FILE "/.mplayer/af_analysis.fifo" /* file name */
#define DEFAULT_PLUGIN "basic_plot"
#define DEFAULT_STRING_SIZE 1024

void usage(char *progname)
{
  char msg[] = 
    "\nUsage:\n\n"
    "      %s [options]\n\n"
    "   Options:\n"
    "      -c, --config=CONFIG_FILE   Path to the config file\n"
    "      -f, --fifo=FIFO_FILE       Path to the af_analysis FIFO\n"
    "      -p, --plugin=PLUGIN        Path or just the name of the plugin to use\n"
    "      --dump_config              Dump default configuration to stdout\n"
    "\n";
  
  if (progname)
    printf(msg,progname);
    
}



int main(int argc, char *argv[])
{
  int i, f_ret;
  char config_file[DEFAULT_STRING_SIZE];
  plugin *plugin_struct;
  /* Configure options *******************************************************/
  config_item_list *cfg_items = NULL;
  char fifo_file[DEFAULT_STRING_SIZE], fifo_file_default[DEFAULT_STRING_SIZE], fifo_file_cmd[DEFAULT_STRING_SIZE];
  char plugin_name[DEFAULT_STRING_SIZE], plugin_name_default[DEFAULT_STRING_SIZE], plugin_name_cmd[DEFAULT_STRING_SIZE];

  add_config_option(&cfg_items, "fifo_file", "Path to the af_analysis FIFO file", STRING, fifo_file, fifo_file_default, NULL, NULL);
  strncpy(fifo_file_default, getenv("HOME"), DEFAULT_STRING_SIZE);
  strncat(fifo_file_default, RELATIVE_FIFO_FILE, DEFAULT_STRING_SIZE - strlen(fifo_file_default));
  strcpy(fifo_file, fifo_file_default);

  add_config_option(&cfg_items, "plugin", "Path to the plugin to use or just it's name", STRING, plugin_name, plugin_name_default, NULL, NULL);
  strncpy(plugin_name, DEFAULT_PLUGIN, DEFAULT_STRING_SIZE);
  strncpy(plugin_name_default, DEFAULT_PLUGIN, DEFAULT_STRING_SIZE);

  strncpy(config_file, getenv("HOME"), DEFAULT_STRING_SIZE);
  strncat(config_file, CONFIG_FILE, DEFAULT_STRING_SIZE - strlen(config_file));
  

  /* Parse command line options **********************************************/
  strcpy(fifo_file_cmd,"");
  strcpy(plugin_name_cmd,"");
  while (1) 
    {      
      int option_index = 0;
      static struct option long_options[] = {
	{ "config",          1, 0, 0 },
	{ "fifo",            1, 0, 0 },
	{ "plugin",          1, 0, 0 },
	{ "dump_config",     0, 0, 0 },
	{ "help",            0, 0, 0 },
	{ 0, 0, 0, 0}
      };
  
      i = getopt_long(argc, argv, "c:f:p:h", long_options, &option_index);
      if (i == -1)
	break;
      
      switch (i) 
	{
	case 0:
	  if (strcmp(long_options[option_index].name,"config") == 0)
	    {
	      if (optarg)
		strncpy(config_file, optarg, DEFAULT_STRING_SIZE);
	    }
	  else if (strcmp(long_options[option_index].name,"fifo") == 0)
	    {
	      if (optarg)
		strncpy(fifo_file_cmd, optarg, DEFAULT_STRING_SIZE);
	    }
	  else if (strcmp(long_options[option_index].name,"plugin") == 0)
	    {
	      if (optarg)
		strncpy(plugin_name_cmd, optarg, DEFAULT_STRING_SIZE);
	    }
	  else if (strcmp(long_options[option_index].name,"dump_config") == 0)
	    {
	      print_config_options(stdout, cfg_items);
	      return 0;
	    }	  
	  else if (strcmp(long_options[option_index].name,"help") == 0)
	    {
	      usage(argv[0]);
	      return 0;
	    }
	  break;
	case 'c': /* -c == --config */
	  if (optarg)
	    strncpy(config_file, optarg, DEFAULT_STRING_SIZE);
	  break;
	case 'f': /* -f == --fifo */
	  if (optarg)
	    strncpy(fifo_file_cmd, optarg, DEFAULT_STRING_SIZE);
	  break;
	case 'p': /* -p == --plugin */
	  if (optarg)
	    strncpy(plugin_name_cmd, optarg, DEFAULT_STRING_SIZE);
	  break;
	case 'h': /* -h == --help */
	  usage(argv[0]);
	  return 0;
	  break;
	default:
	  break;
	}
    }

#ifdef DEBUG
  printf("Reading config from file '%s'\n",config_file);
#endif
  /* Parse config file *******************************************************/
  f_ret = parse_config_file(config_file,cfg_items);
  if (f_ret == -1)
    {
      fprintf(stderr, "ERROR: Internal error, passed null pointer parameter to function(%s). Please report this bug!\n","parse_config_file()");
    }
  else if (f_ret == -2)
    {
      fprintf(stderr, "ERROR: The config file doesn't exist: \"%s\"\n", config_file);
    }

  /* override option from file with the command line ones ********************/
  if (strlen(fifo_file_cmd) > 0)
    strncpy(fifo_file,fifo_file_cmd, DEFAULT_STRING_SIZE);
  if (strlen(plugin_name_cmd) > 0)
    strncpy(plugin_name,plugin_name_cmd, DEFAULT_STRING_SIZE);

  /* Load the plugin *********************************************************/
  plugin_struct = (plugin *) calloc(1, sizeof(plugin));
  f_ret = load_plugin(plugin_name,&plugin_struct);
  if (f_ret == -1)
    {
      fprintf(stderr, "ERROR: Internal error, passed null pointer parameter to function(%s). Please report this bug!\n","load_plugin()");
    }
  else if (f_ret == -2)
    {
      fprintf(stderr, "ERROR: Could not load plugin '%s'\n", plugin_name);
    }
  else if (f_ret == -3)
    {
      fprintf(stderr, "ERROR: Could not init the plugin '%s'\n", plugin_name);
    }

  if (f_ret == 0)
    {
      /* Print info about the plugin *********************************************/
      fprintf(stdout, 
	      "Plugin Info:\n"
	      "   Name.......: %s\n"
	      "   Description: %s\n"
	      "   Version....: %s\n"
	      "   Author.....: %s\n"
	      "   URL........: %s\n",
	      (plugin_struct->info.name != NULL ? plugin_struct->info.name : "<Not Avaiable>"),
	      (plugin_struct->info.description != NULL ? plugin_struct->info.description : "<Not Avaiable>"),
	      (plugin_struct->info.version != NULL ? plugin_struct->info.version : "<Not Avaiable>"),
	      (plugin_struct->info.author != NULL ? plugin_struct->info.author : "<Not Avaiable>"),
	      (plugin_struct->info.url != NULL ? plugin_struct->info.url : "<Not Avaiable>")
	      );


      /* Enter the main loop (the real work) *************************************/
      f_ret = main_loop(cfg_items, plugin_struct);
      if (f_ret == -1)
	{
	  fprintf(stderr, "ERROR: Internal error, passed null pointer parameter to function(%s). Please report this bug!\n","main_loop()");
	}
      else if (f_ret == -2)
	{
	  fprintf(stderr, "ERROR: Could not open fifo '%s'\n", (char *)(* get_config_item("fifo_filename",cfg_items)).value);
	}
      else if (f_ret == -3)
	{
	  fprintf(stderr, "ERROR: Could not init SDL\n");
	}
      else if (f_ret == -4)
	{
	  fprintf(stderr, "ERROR: Could not create SDL_Surface\n");
	}
    }

  /* Unload the plugin ******************************************************/
  unload_plugin(plugin_struct);
  if (plugin_struct) free(plugin_struct);

  /* Free the cfg_items *****************************************************/
  free_config_option_list(&cfg_items);

  return 0;
}
