#include "read_config.h"

int parse_int(const char *string, int *value, const int *min, const int *max)
{
  if ((value == NULL) || (string == NULL))
    return PARSE_ERR_NULL_POINTER;

  sscanf(string,"%i",value);

  if ( ( ( max != NULL ) && ( (*value) > (*max) ) ) ||
       ( ( min != NULL ) && ( (*value) < (*min) ) )
       )
       return PARSE_ERR_OUT_OF_RANGE;

  return 0;
}

int parse_float(const char *string, float *value, const float *min, const float *max)
{
  if ((value == NULL) || (string == NULL))
    return PARSE_ERR_NULL_POINTER;

  sscanf(string,"%f",value);

  if ( ( ( max != NULL ) && ( (*value) > (*max) ) ) ||
       ( ( min != NULL ) && ( (*value) < (*min) ) )
       )
       return PARSE_ERR_OUT_OF_RANGE;

  return 0;
}

int parse_bool(const char *string, int *value)
{
  char tmp[255];
  
  if ((value == NULL) || (string == NULL))
    return PARSE_ERR_NULL_POINTER;

  (*value) = 0;
  if (sscanf(string,"%i",value) < 1)
    {
      if (parse_string(string, tmp, 255) < 0)
	return PARSE_ERR_INVALID_VALUE;
      else
	{
	  to_lowercase(tmp,tmp,strlen(tmp));
	  if ( (strcoll(tmp,"true") == 0) ||
	       (strcoll(tmp,"t") == 0) ||
	       (strcoll(tmp,"on") == 0) || 
	       (strcoll(tmp,"y") == 0)
	       ) 
	    {
	      (*value) = 1;
	    }
	  else if ( (strcoll(tmp,"false") == 0) ||
		    (strcoll(tmp,"f") == 0) ||
		    (strcoll(tmp,"off") == 0) || 
		    (strcoll(tmp,"n") == 0)
		    ) 
	    {
	      (*value) = 0;
	    }
	  else
	    return PARSE_ERR_INVALID_VALUE;
	}
    }
  else
    {
      if (((*value) != 0) && ((*value) != 1))
	return PARSE_ERR_INVALID_VALUE;
    }

  return 0;
}

int parse_string(const char *string, char *value, int size)
{
  int i=0,j=0;
  char quotes[255];
  int pos_quotes=0;
  
  if ((string==NULL) || (value==NULL))
    {
      return PARSE_ERR_NULL_POINTER;
    }


  /* skip initial spaces */
  for (i=0; i < size; i++)
    if (string[i] != ' ')
      break;

  for (j=0; i < size; i++)
    {
      if ((string[i] == '\0') || (string[i] == '\n'))
	break;
      else
	{
	  if ((string[i] == '"') || (string[i] == '\''))
	    {
	      if (j == 0)
		{
		  /* is first quote and no text yet?*/
		  quotes[pos_quotes++] = string[i];
		}
	      else if (pos_quotes > 0)
		{
		  /* this is a 'closing' quote ? */
		  if (quotes[pos_quotes-1] == string[i])
		    {
		      pos_quotes--;
		      if (pos_quotes > 0)
			value[j++] = string[i];			
		    }
		  else
		    {
		      quotes[pos_quotes++] = string[i];
		      value[j++] = string[i];
		    }
		}
	    }
	  else
	    {
	      value[j++] = string[i];
	    }
	}
    }
  value[j] = '\0';

  trim(value);

  if (pos_quotes > 0)
    return PARSE_ERR_INVALID_VALUE;
  else
    return 0;
}


int parse_color(const char *string, int *value)
{
  char *idx = NULL, *string2 = NULL;
  int R=0, G=0, B=0, f_ret=0; 
  int min = 0, max = 255;

  if ((value == NULL) || (string == NULL))
    return PARSE_ERR_NULL_POINTER;
  
  string2 = (char *) calloc(strlen(string), sizeof(char));
  strcpy(string2, string);

  (*value) = 0;
  idx = string2;
  if ((f_ret = parse_int(idx, &R, &min, &max)) < 0)
    {
      free(string2);
      return f_ret;
    }
  idx = index(idx, ',')+1;
  if ((f_ret = parse_int(idx, &G, &min, &max)) < 0)
    {
      free(string2);
      return f_ret;
    }
  idx = index(idx, ',')+1;
  if ((f_ret = parse_int(idx, &B, &min, &max)) < 0)
    {
      free(string2);
      return f_ret;
    }

  (*value) =  R << 16;
  (*value) += G << 8;
  (*value) += B;
  
  if (string2) free(string2);

  return 0;
}




int to_lowercase(const char *src, char *dst, int size)
{
  int i=0;

  if ((src == NULL) || (dst == NULL))
    return -1;

  for (i=0; i < size; i++)
    {
      if (src[i] == '\0')
	break;
      else if ( (src[i] >= 'A') && (src[i] <= 'Z') )
	{
	  dst[i] = src[i] - 'A' + 'a';
	}
      else
	{
	  dst[i] = src[i];
	}
    }
  dst[i] = '\0';
  return 0;
}

int trim(char *string)
{
  int i,j,k, len;
  if (string == NULL)
    return -1;

  i = 0;
  j = 0;
  k = 0;
  len = strlen(string);
  for (i=0; (string[i] == ' ') && (i < len); i++);
  for (j=len-1; (string[j] == ' ') && (j > i); j--);
  for (k=0; i <= j;  string[k++] = string[i++]);
  string[k]='\0';

  return 0;
}


int split_pair(const char *string, char *key, char *value)
{
  int i, j;
  char *idx;

  if ( (string == NULL) || (key == NULL) || (value == NULL) )
    return -1;
  
  idx = index(string, '=');

  if (idx == NULL)
    return -2;

  i=j=0;
  while ( string[i] != '=' )
    key[j++] = string[i++];
  key[j] = '\0';
  
  i++;
  j=0;
  while ( string[i] != '\0' )
    value[j++] = string[i++];
  value[j] = '\0';

  return 0;
}



int parse_config_file(const char *file, config_item_list *items)
{
  FILE *fp = NULL;
  int i,j, is_comment=0, is_text_begin=0, found=0, line_number=0, f_ret=0;
  char line[LINE_SIZE], option_name[LINE_SIZE], option_value[LINE_SIZE], tmp_c;
  char *tmp_cp;
  config_item_list *p;

  if ((items == NULL) || (file == NULL))
    return -1;
  
  if ((fp = fopen(file, "r")) == NULL)
    return -2;

  line_number = 1;
  while (! feof(fp))
    {
      is_comment = 0;
      is_text_begin = 0;
      j=0;
      /* read a line */
      for (i=0; i < LINE_SIZE-1; i++)
	{
	  tmp_c = getc(fp);

	  if (tmp_c == EOF)
	    break;

	  if ((tmp_c == '\0') || (tmp_c == '\n'))
	    break; /* line read */
	  else if (tmp_c == '#')
	    is_comment = 1;
	  else if ((tmp_c != ' ') || (is_text_begin == 1))
	    {
	      if (! is_comment) {
		line[j++] = tmp_c;
		is_text_begin = 1;
	      }
	    }
	}
      line[j] = '\0';

      if ( strlen(line) > 0 )
	{
	  strcpy(option_value,"");
	  if (split_pair(line, option_name, option_value) < 0)
	    {
	      fprintf(stderr,"ERROR: Config file '%s', line %d: '%s' wrong line format! Lines should be key = value\n",file, line_number, line);
	    }	  
	  else
	    {
	      found = 0;
	      p = items;
	      while (p)
		{
		  trim(option_name);
		  tmp_cp = strstr(option_name, p->item.name);
		  if (tmp_cp != NULL)
		    {
		      found = 1;
		      f_ret = 0;
		      switch (p->item.option_type) 
			{
			case INTEGER:
			  f_ret = parse_int(option_value,(int *) p->item.value, (int *) p->item.min, (int *) p->item.max);
			  if (f_ret < 0)
			    (*(int *)(p->item.value)) = (*(int *)(p->item.default_value));
			  break;
			  
			case FLOAT:
			  f_ret = parse_float(option_value,(float *) p->item.value, (float *) p->item.min, (float *) p->item.max);
			  if (f_ret < 0)
			    (*(float *)(p->item.value)) = (*(float *)(p->item.default_value));
			  break;
			  
			case BOOLEAN:
			  f_ret = parse_bool(option_value,(int *) p->item.value);
			  if (f_ret < 0)
			    (*(int *)(p->item.value)) = (*(int *)(p->item.default_value));
			  break;
			  
			case STRING:
			  f_ret = parse_string(option_value, (char *) p->item.value, LINE_SIZE);
			  if (f_ret < 0)
			    strncpy((char *)p->item.value, (char *)p->item.default_value, LINE_SIZE);
			  break;
			  
			case COLOR:
			  f_ret = parse_color(option_value,(int *) p->item.value);
			  if (f_ret < 0)
			    (*(int *)(p->item.value)) = (*(int *)(p->item.default_value));
			  break;
			  
			default:
			  break;
			}
		      
		      if (f_ret != 0)
			{
			  fprintf(stderr,"ERROR: Parsing config file '%s', line %d: ",file, line_number);
			  switch (f_ret) 
			    {
			    case PARSE_ERR_NULL_POINTER:
			      fprintf(stderr,"Null pointer error! (This souldn't happer -- Internal error, please report the bug!)");
			      break;
			    case PARSE_ERR_INVALID_VALUE:
			      fprintf(stderr,"Invalid value: '%s'", option_value);
			      break;
			    case PARSE_ERR_OUT_OF_RANGE:
			      fprintf(stderr,"Value out of range: ");
			      switch (p->item.option_type)
				{
				case INTEGER:
				  fprintf(stderr,"%d Should be between %d and %d", *(int *)p->item.value, *(int *)p->item.min, *(int *)p->item.max);
				  break;
				case FLOAT:
				  fprintf(stderr,"%f Should be between %f and %f", *(float *)p->item.value, *(float *)p->item.min, *(float *)p->item.max);
				  break;
				case COLOR:
				  fprintf(stderr,"%s Each value should be between %d and %d",option_value, 0, 255);
				  break;
				default:
				  fprintf(stderr,"%s", option_value);
				  break;
				}			  
			    }
			  fprintf(stderr,"  (Please read the manual)\n");
			}
		    }
		  p = p->next;
		}
	      if (! found)
		{
		  fprintf(stderr,"ERROR: Config file '%s', line %d: Option doesn't recognized: '%s'\n",file, line_number, option_name); 
		}
	    }
	}
      line_number++;
    }
  
  fclose(fp);

  return 0;
}


config_item *get_config_item(const char *name, config_item_list *items)
{
  config_item_list *p;
  config_item *item = NULL;

  if (name == NULL)
    return NULL;

  p = items;
  while (p)
    {
      if (strcmp(p->item.name,name)==0)
	{
	  item = &(p->item);
	  break;
	}
      p = p->next;
    }
  
  return item;
}


void print_config_options(FILE *fp, config_item_list *items)
{
  config_item_list *p = items;
  time_t t = time(NULL);

  fprintf(fp,"# audio_visualization configuration file  --- %s\n",ctime(&t));
  while (p)
    {
      fprintf(fp, "# %s: %s. ", p->item.name, p->item.description);
      switch (p->item.option_type)
	{
	case INTEGER:
	  fprintf(fp,"(Value type: Integer; Default value: %d; Minumum: %d; Maximum: %d)\n",
		  *(int *)p->item.default_value, *(int *)p->item.min, *(int *)p->item.max);
	  break;
	case FLOAT:
	  fprintf(fp,"(Value type: Float; Default value: %f; Minumum: %f; Maximum: %f)\n",
		  *(float *)p->item.default_value, *(float *)p->item.min, *(float *)p->item.max);
	  break;
	case BOOLEAN:
	  fprintf(fp,"(Value type: Boolean; Default value: %d. Possible values: true,t,y,on,1,false,f,n,off,0)\n",
		  *(int *)p->item.default_value);
	  break;
	case STRING:
	  fprintf(fp,"(Value type: String; Default value: \"%s\")\n",
		  (char *)p->item.default_value);
	  break;
	case COLOR:
	  fprintf(fp,"(Value type: Color; Default value: %d,%d,%d; Each value should be between 0 and 255)\n",
		  (*(int *)p->item.default_value) >> 16, 
		  ((*(int *)p->item.default_value) % 65536) >> 8,
		  ((*(int *)p->item.default_value) % 256)
		  );
	  break;
	default:
	  break;       
	}
      
      fprintf(fp, "%s=", p->item.name);
      switch (p->item.option_type)
	{
	case INTEGER:
	  fprintf(fp,"%d # Default: %d; Minumum: %d; Maximum: %d\n",
		  *(int *)p->item.value, *(int *)p->item.default_value, *(int *)p->item.min, *(int *)p->item.max);
	  break;
	case FLOAT:
	  fprintf(fp,"%f # Default: %f; Minumum: %f; Maximum: %f\n",
		  *(float *)p->item.value, *(float *)p->item.default_value, *(float *)p->item.min, *(float *)p->item.max);
	  break;
	case BOOLEAN:
	  fprintf(fp,"%d # Default: %d. Possible values: true,t,y,on,1,false,f,n,off,0\n",
		  *(int *)p->item.value, *(int *)p->item.default_value);
	  break;
	case STRING:
	  fprintf(fp,"\"%s\" # Default: \"%s\"\n",
		  (char *)p->item.value, (char *)p->item.default_value);
	  break;
	case COLOR:
	  fprintf(fp,"%d,%d,%d # Default value: %d,%d,%d; Each value should be between 0 and 255\n",
		  (*(int *)p->item.value) >> 16,
		  ((*(int *)p->item.value) % 65536) >> 8,
		  ((*(int *)p->item.value) % 256),
		  (*(int *)p->item.default_value) >> 16,
		  ((*(int *)p->item.default_value) % 65536) >> 8,
		  ((*(int *)p->item.default_value) % 256)
		  );
	  break;
	default:
	  fprintf(fp,"# ERROR GENERATING THIS FILE: Option '%s', type='%d'. Please report this bug!\n",
		  p->item.name, p->item.option_type);
	  break;       
	}
      fprintf(fp,"\n");

      p = p->next;
    }
}

int add_config_option(config_item_list **items, 
		      const char *name, const char *description,
		      option_type_enum option_type, 
		      void *value, void *default_value, 
		      void *min, void *max)
{
  config_item_list *p;
  if (items == NULL)
    return -1;

  p = (*items);
  (*items) = (config_item_list *) calloc(1, sizeof(config_item_list));
  if ((*items) == NULL)
    return -1;

  (*items)->next = p;

  (*items)->item.name = (char *) calloc(strlen(name), sizeof(char));
  (*items)->item.description = (char *) calloc(strlen(description), sizeof(char));
  (*items)->item.option_type = option_type;
  (*items)->item.value = value;
  (*items)->item.default_value = default_value;
  (*items)->item.min = min;
  (*items)->item.max = max;
  strcpy((*items)->item.name, name);
  strcpy((*items)->item.description, description);
  
  return 0;
}

void free_config_option_list(config_item_list **items)
{
  config_item_list *p = NULL, *aux =NULL;

  if (items == NULL)
    return;

  p = *items;
  while (p)
    {
      aux = p;
      if (p->item.name) free(p->item.name);
      if (p->item.description) free(p->item.description);
      p = p->next;
      free(aux);
    }
  
  *items = NULL;
}
