#ifndef READ_CONFIG_H
#define READ_CONFIG_H

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <time.h>

#define LINE_SIZE 1024

typedef enum { STRING, BOOLEAN, INTEGER, FLOAT , COLOR } option_type_enum;


#define PARSE_ERR_NULL_POINTER  -1
#define PARSE_ERR_INVALID_VALUE -2
#define PARSE_ERR_OUT_OF_RANGE  -3


typedef struct config_item_s {
  char *name;
  char *description;
  option_type_enum option_type;
  void *value;
  void *default_value;
  void *min; /* should accept the >, <, =, ... operations ! */
  void *max; /* should accept the >, <, =, ... operations ! */ 
} config_item;

typedef struct config_item_list_s {
  config_item item;
  struct config_item_list_s *next;
} config_item_list;

/** 
 * parse_int: parse integer in string
 * @param string the string which contains the value to be parsed
 * @param value where the parsed value will be returned
 * @return -1 = error occured, 0 = ok
 */
int parse_int(const char *string, int *value, const int *min, const int *max);

/** 
 * parse_float: parse float in string
 * @param string the string which contains the value to be parsed
 * @param value where the parsed value will be returned
 * @return -1 = error occured, 0 = ok
 */
int parse_float(const char *string, float *value, const float *min, const float *max);

/** 
 * parse_string: parse string (between "" or '') in string
 * @param string the string which contains the value to be parsed
 * @param value where the parsed value will be returned
 * @param size is the size of value
 * @return -1 = error occured, 0 = ok
 */
int parse_string(const char *string, char *value, int size);

/** 
 * parse_bool: parse boolean (true==t==y==on==1, false==f==n==off=0) in string
 * @param string the string which contains the value to be parsed
 * @param value where the parsed value will be returned
 * @return -1 = error occured, 0 = ok
 */
int parse_bool(const char *string, int *value);


/** 
 * parse_color: parse color R,G,B (0-255 each) in string
 * @param string the string which contains the value to be parsed
 * @param value where the parsed value will be returned
 * @return -1 = error occured, 0 = ok
 */
int parse_color(const char *string, int *value);


/** 
 * to_lowercase: change the text to lowercase 
 * @param src source string
 * @param dst destination string
 * @param size size of the strings
 * @return -1 = error, 0 = ok
 */
int to_lowercase(const char *src, char *dst, int size);


/**
 * trim: cut off spaces from begining and end of the string
 * @param string string to be trimmed
 * @return -1 = error, 0 = ok
 */
int trim(char *string);

/**
 * split_pair: split the string "key = value"
 * @param string the string that contains the pair
 * @param key where to return the key
 * @param value where to return the value
 * @return -1 = error, 0 = ok
 */
int split_pair(const char *string, char *key, char *value);


/**
 * parse_config_file: parse the config files accordinly to the list of itens gived 
 * @param file the path to config file
 * @param items the list with items
 * @return 0=ok, -1 = null pointer parameter, -2 = error opening file
 */
int parse_config_file(const char *file, config_item_list *items);

/**
 * get_config_item: get the config item named 'name'
 * @param name the item's name
 * @param items the list with items
 * @return NULL=if not found or the pointer to the config_item found
 */
config_item *get_config_item(const char *name, config_item_list *items);


/**
 * print_config_options: print the options to the fd gived
 * @param fp file pointer, where the data should be printed
 * @param items the list with items
 */
void print_config_options(FILE *fp, config_item_list *items);


/**
 * add_config_option: add the item to the list
 * @param items the list with items
 * @param name the name of the item
 * @param description the description of the item
 * @param option_type the type of the item
 * @param value pointer to the value of the item
 * @param default_value pointer to the item's default value
 * @param min pointer to the item's minimum value
 * @param max pointer to the item's maximum value
 * @return -1 = error, 0 = ok
 */
int add_config_option(config_item_list **items, 
		      const char *name, const char *description,
		      option_type_enum option_type, 
		      void *value, void *default_value, 
		      void *min, void *max);

/**
 * free_config_option_list: free the list
 * Warning: The name and description fields from the (*items)->item record *WILL* be freed too!
 * @param items the list with items.
 */
void free_config_option_list(config_item_list **items);

#endif /* READ_CONFIG_H */
